@@ -65,20 +65,19 @@ MVC::Neaf->route( cgi => $script => sub {
     my $req = shift;
 
     $req->redirect( $req->path . "/and/beyond" )
-        unless $req->path_info( '.+' );
+        unless $req->path_info( );
 
     my @error;
     local $SIG{__DIE__} = sub { push @error, shift };
     return {
         -template => \$tpl,
         error => \@error,
-        path_info => $req->path_info( '.*' ),
         map {
             $_ => eval { $req->$_ } || "unimplemented: $_";
         } qw(method path http_version scheme hostname port
-            script_name client_ip referer user_agent),
+            script_name path_info client_ip referer user_agent),
     };
-}, description => $descr );
+}, description => $descr, path_info_regex => '.*' );
 
 MVC::Neaf->alias( "/request/parser" => "/cgi/$script" );
 
