@@ -4,33 +4,29 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = 0.1302;
+our $VERSION = 0.14;
 
 =head1 NAME
 
-MVC::Neaf - Not Even A Framework for very simple web apps
+MVC::Neaf - Not Even A (Web Application) Framework
 
 =head1 OVERVIEW
 
-Neaf [ni:f] stands for Not Even An (MVC) Framework.
+Neaf [ni:f] stands for Not Even A Framework.
 
-It is made for lazy people without an IDE.
-
-The Model is assumed to be just a regular Perl module,
+The B<Model> is assumed to be just a regular Perl module,
 no restrictions are imposed on it.
 
-The View is an object with one method, C<render>, receiving a hashref
+The B<View> is an object with one method, C<render>, receiving a hashref
 and returning rendered content as string plus optional content-type header.
 
-The Controller is reduced to just one function which receives
-a <MVC::Neaf::Request> object containing all it needs to know
-and returns a simple \%hashref which is forwarded to View.
+The B<Controller> is broken down into handlers associated with URI paths.
+Each such handler receives a L<MVC::Neaf::Request> object
+containing all it needs to know about the outside world,
+and returns a simple C<\%hashref> which is forwarded to View.
 
-Multiple controllers can be set up in the same application
-under different paths.
-
-Please see the examples directory in this distribution
-which demonstrates the features of Neaf.
+Please see the C<example> directory in this distribution
+that demonstrates the features of Neaf.
 
 =head1 SYNOPSIS
 
@@ -56,16 +52,15 @@ as a CGI script, PSGI application, or Apache handler.
 
 =head2 THE CONTROLLER
 
-The Controller sub receives an L<MVC::Neaf::Request> object
-and outputs a \%hashref.
+The handler sub receives an L<MVC::Neaf::Request> object
+and outputs a C<\%hashref>.
 
 It may also die, which will be interpreted as an error 500,
 UNLESS error message starts with 3 digits and a whitespace,
 in which case this is considered the return status.
-E.g. die 404; is a valid method to return "Not Found" right away.
+E.g. C<die 404;> is a valid method to return "Not Found" right away.
 
-Much like in L<Dancer> or L<Kelp>, multiple controllers may be configured
-under different paths using the C<route( path =E<gt> CODEREF );>
+Handlers are set using the C<route( path =E<gt> CODEREF );>
 method discussed below.
 
 =head2 THE REQUEST
@@ -96,19 +91,18 @@ Just use qr/.*/ if you know better.
 Also there are some methods that affect the reply,
 mainly the headers, like C<set_cookie> or C<redirect>.
 This is a step towards a know-it-all God object,
-however, mapping those proverties into a hashref turned out to be
+however, mapping those properties into a hashref turned out to be
 too cumbersome.
 
 =head2 THE RESPONSE
 
 B<The response> may contain regular keys, typically alphanumeric,
-as well as a predefined set of dash-prefixed keys which control
-app's behaviour.
+as well as a predefined set of dash-prefixed keys to control
+Neaf itself.
 
 I<-Note -that -dash-prefixed -options -look -antique
 even to the author of this writing.
-They smell like Tk and CGI. They feel so 90's!
-However, it looks like a meaningful and B<visible> way to separate
+However, it is a concise and B<visible> way to separate
 auxiliary parameters from users's data,
 without requiring a more complex return structure
 (two hashes, array of arrays etc).>
@@ -120,15 +114,15 @@ The small but growing list of these -options is as follows:
 =item * -content - Return raw data and skip view processing.
 E.g. display generated image.
 
-=item * -continue - A callback which receives the Request object.
-It will be executed after the headers and pre-generated content
-are served to the client, and may use $req->write( $data )
-and $req->close to write more data.
+=item * -continue - A callback that receives the Request object.
+It will be executed AFTER the headers and pre-generated content
+are served to the client, and may use C<$req-E<gt>write( $data );>
+and C<$req-E<gt>close;> to output more data.
 
 =item * -jsonp - Used by JS view module as a callback name to produce a
 L<jsonp|https://en.wikipedia.org/wiki/JSONP> response.
-Callback is ignored unless it is a set of identifiers separated by dots,
-for security reasons.
+Callback MUST be a set of identifiers separated by dots.
+Otherwise it's ignored for security reasons.
 
 =item * -location - HTTP Location: header.
 
@@ -146,7 +140,7 @@ Views are initialized lazily and cached by the framework.
 C<TT>, C<JS>, C<Full::Module::Name>, and C<$view_predefined_object>
 are currently supported.
 New short aliases may be created by
-C<MVC::Neaf-E<gt>load_view( "name" => $your_view );>. (See below).
+C<MVC::Neaf-E<gt>load_view( "name" =E<gt> $your_view );> (see below).
 
 =back
 
@@ -168,9 +162,12 @@ use Encode;
 use POSIX qw(strftime);
 use parent qw(Exporter);
 
-our @EXPORT_OK = qw(neaf_err);
+our @EXPORT_OK = qw(neaf_err neaf get post);
+our %EXPORT_TAGS = (
+    sugar => [qw[neaf get post]],
+);
 
-use MVC::Neaf::Util qw(http_date);
+use MVC::Neaf::Util qw(http_date canonize_path path_prefixes);
 use MVC::Neaf::Request;
 
 our $Inst = __PACKAGE__->new;
@@ -198,7 +195,7 @@ Exactly one leading slash will be prepended no matter what you do.
 =item * method - list of allowed HTTP methods.
 Default is [GET, POST, HEAD].
 Multiple handles can be defined for the same path, provided that
-metyhods do not intersect.
+methods do not intersect.
 
 =item * path_info_regex => qr/.../ - allow URI subpaths matching
 the specified expression to be handled by this handler.
@@ -206,20 +203,25 @@ the specified expression to be handled by this handler.
 If specified, 404 error will be returned unless PATH_INFO matches the regex
 (without the leading slash).
 
-Starting from v.0.16, this parameter is going to be REQUIRED for path_info()
+Starting from v.0.16, this parameter is going to be REQUIRED for C<path_info()>
 method to be used in the handler.
 Until then, a DEPRECATED warning will be generated for a naked path_info call.
 
 B<EXPERIMENTAL>. Name and semantics MAY change in the future.
 
 =item * view - default View object for this Controller.
-Must be an object with a C<render> methods, or a CODEREF
-receiving hash and returning a list of two scalars.
+Must be an object with a C<render> method, or a CODEREF
+receiving hashref and returning a list of two scalars
+(content and content-type).
 
 =item * cache_ttl - if set, set Expires: HTTP header accordingly.
 
 B<EXPERIMENTAL>. Name and semantics MAY change in the future.
 
+=item * default - a C<\%hash> with default values for handler's return value.
+
+B<EXPERIMENTAL>. Name and semantics MAY change in the future.
+
 =item * description - just for information, has no action on execution.
 This will be displayed if application called with --list (see L<MVC::Neaf::CLI>).
 
@@ -228,6 +230,9 @@ This will be displayed if application called with --list (see L<MVC::Neaf::CLI>)
 =cut
 
 my $year = 365 * 24 * 60 * 60;
+my %known_route_args;
+$known_route_args{$_}++ for qw(default method view cache_ttl path_info_regex description);
+
 sub route {
     my $self = shift;
 
@@ -243,13 +248,14 @@ sub route {
     my (%args) = @_;
     $self = $Inst unless ref $self;
 
-    # Sanitize path so that we have exactly one leading slash
-    # root becomes nothing (which is OK with us).
-    $path =~ s#^/*#/#;
-    $path =~ s#/+$##;
+    # kill extra args
+    my @extra = grep { !$known_route_args{$_} } keys %args;
+    $self->_croak( "Unexpected keys in route setup: @extra" )
+        if @extra;
 
-    $args{method} ||= [qw[ GET POST HEAD ]];
-    $args{method} = [ $args{method} ] unless ref $args{method} eq 'ARRAY';
+    $path = canonize_path( $path );
+
+    _listify( \$args{method}, qw( GET POST HEAD ) );
 
     my @dupe = grep { exists $self->{route}{$path}{$_} } @{ $args{method} };
     $self->_croak( "Attempting to set duplicate handler for [@dupe] "
@@ -285,6 +291,12 @@ sub route {
         $profile{view} = $view;
     };
 
+    if (my $def = $args{default}) {
+        $self->_croak( "default must be unblessed hash" )
+            unless ref $def eq 'HASH';
+        $profile{todo_default} = $args{default};
+    };
+
     if ( $args{cache_ttl} ) {
         $self->_croak("cache_ttl must be a number")
             unless looks_like_number($args{cache_ttl});
@@ -294,9 +306,9 @@ sub route {
         $profile{cache_ttl} = $args{cache_ttl};
     };
 
-    # ready, install handler & burn cache
+    # ready, shallow copy handler & burn cache
     delete $self->{route_re};
-    $self->{route}{ $path }{$_} = \%profile
+    $self->{route}{ $path }{$_} = { %profile, my_method => $_ }
         for @{ $args{method} };
 
     return $self;
@@ -316,15 +328,12 @@ sub alias {
     my ($self, $new, $old) = @_;
     $self = $Inst unless ref $self;
 
-    $old =~ s#^/*#/#;
-    $old =~ s#/+$##;
+    $new = canonize_path( $new );
+    $old = canonize_path( $old );
 
     $self->{route}{$old}
         or $self->_croak( "Cannot create alias for unknown route $old" );
 
-    # The same sanitizing as in route
-    $new =~ s#^/*#/#;
-    $new =~ s#/+$##;
     $self->_croak( "Attempting to set duplicate handler for path "
         .( length $new ? $new : "/" ) )
             if $self->{route}{ $new };
@@ -338,11 +347,11 @@ sub alias {
 
 =head2 static( $req_path => $file_path, %options )
 
-Serve static content located under $file_path.
+Serve static content located under C<$file_path>.
 
 File type detection is based on extention.
 This MAY change in the future.
-Known file types are listed in %MVC::Neaf::X::Files::ExtType hash.
+Known file types are listed in C<%MVC::Neaf::X::Files::ExtType> hash.
 Patches welcome.
 
 %options may include:
@@ -416,31 +425,31 @@ sub pre_route {
 
 =head2 load_view( $view_name, [ $object || $module_name, %options ] )
 
-Return a cached View object named $view_name.
+Fetch existing, if any, or set up a new view object named C<$view_name>.
+
 This will be called whenever the app returns { -view => "foo" }.
 
-If no such object was found, attempts to add new object to cache:
+If no view was found, the following rules apply:
 
 =over
 
-=item * if object is given, just saves it.
+=item * if object is given, just save it.
 
-=item * if module name + parameters is given, tries to load module
+=item * if module name + parameters is given, try to load module
 and create new() instance.
 
-=item * as a last resort, loads stock view: C<TT>, C<JS>, or C<Dumper>.
+=item * as a last resort, load stock view: C<TT>, C<JS>, or C<Dumper>.
 Those are prefixed with C<MVC::Neaf::View::>.
 
 =back
 
-If set_forced_view was called with an argument,
-return THAT view instance instead of all above.
+If C<set_forced_view> was called, return its argument instead.
 
 So the intended usage is as follows:
 
     # in the app initialisation section
     # this can be omitted when using TT, JS, or Dumper as view.
-    MVC::Neaf->load_view( foo => TT, INCLUDE_PATH => "/foo/bar" );
+    MVC::Neaf->load_view( foo => 'TT', INCLUDE_PATH => "/foo/bar" );
         # Short alias for MVC::Neaf::View::TT
 
     # in the app itself
@@ -502,13 +511,47 @@ Controller return always overrides these values.
 
 Returns self.
 
+B<DEPRECATED>. Use C<MVC::Neaf-E<gt>set_path_defaults( '/', { ... } );> instead.
+
 =cut
 
 sub set_default {
     my ($self, %data) = @_;
     $self = $Inst unless ref $self;
 
-    $self->{defaults}{$_} = $data{$_} for keys %data;
+    carp "DEPRECATED use set_path_defaults( '/', \%data ) instead of MVC::Neaf->set_default()";
+
+    return $self->set_path_defaults( '/', \%data );
+};
+
+=head2 set_path_defaults ( '/path' => \%values )
+
+Use given values as defaults for ANY handler below given path.
+A value of '/' means global.
+
+Longer paths override shorter ones; route-specific defaults override these;
+and anything defined inside handler takes over once again.
+
+B<EXPERIMENTAL> Name and meaning MAY change in the future.
+
+=cut
+
+sub set_path_defaults {
+    my ($self, $path, $src) = @_;
+    $self = $Inst unless ref $self;
+
+    $self->_croak("arguments must be a scalar and a hashref")
+        unless defined $path and !ref $path and ref $src eq 'HASH';
+
+    # canonize path
+    # CANONIZE
+    $path =~ s#/+#/#;
+    $path =~ s#^/*#/#;
+    $path =~ s#/$##;
+    my $dst = $self->{path_defaults}{$path} ||= {};
+    $dst->{$_} = $src->{$_}
+        for keys %$src;
+
     return $self;
 };
 
@@ -516,8 +559,8 @@ sub set_default {
 
 Set a handler for managing sessions.
 
-If such handler is set, the request object will provide session(),
-save_session(), and delete_session() methods to manage
+If such handler is set, the request object will provide C<session()>,
+C<save_session()>, and C<delete_session()> methods to manage
 cross-request user data.
 
 % options may include:
@@ -596,7 +639,7 @@ The following options will be passed to coderef:
 
 =item * status - status being returned (500 in case of 'view');
 
-=item * caller - array with the point where MVC::Neaf->route was set up;
+=item * caller - array with the point where C<MVC::Neaf-E<gt>route> was set up;
 
 =item * error - exception, if there was one.
 
@@ -604,12 +647,12 @@ The following options will be passed to coderef:
 
 The coderef MUST return an unblessed hash just like controller does.
 
-In case of exception or unexpected return format text message "Error XXX"
+In case of exception or unexpected return format text message "Error NNN"
 will be returned instead.
 
 =head2 set_error_handler ( status => \%hash )
 
-Return a static template as { %options, %hash }.
+Return a static template as C<{ %options, %hash }>.
 
 =cut
 
@@ -651,7 +694,7 @@ sub error_template {
 =head2 on_error( sub { my ($req, $err) = @_ } )
 
 Install custom error handler for dying controller.
-Neaf's own exceptions and 'die \d\d\d' status returns will NOT
+Neaf's own exceptions and C<die \d\d\d> status returns will NOT
 trigger it.
 
 E.g. write to log, or something.
@@ -679,7 +722,7 @@ sub on_error {
 =head2 run()
 
 Run the applicaton.
-This should be in the last statement in your appication main file.
+This should be the last statement in your appication main file.
 
 If called in void context, assumes CGI is being used and instantiates
 L<MVC::Neaf::Request::CGI>.
@@ -689,7 +732,7 @@ enters debug mode via L<MVC::Neaf::CLI>.
 Otherwise returns a PSGI-compliant coderef.
 This will also happen if you application is C<require>'d,
 meaning that it returns a true value and actually serves nothing until
-run() is called again.
+C<run()> is called again.
 
 Running under mod_perl requires setting a handler with
 L<MVC::Neaf::Request::Apache2>.
@@ -747,7 +790,7 @@ Currently only one function is exportable:
 
 Rethrow Neaf's internal exceptions immediately, do nothing otherwise.
 
-If no argument if given, acts on current $@ value.
+If no argument if given, acts on current C<$@> value.
 
 Currently Neaf uses exception mechanism for internal signalling,
 so this function may be of use if there's a lot of eval blocks
@@ -776,6 +819,248 @@ sub neaf_err(;$) { ## no critic # prototype it for less typing on user's part
     die $err;
 };
 
+=head1 EXPERIMENTAL FUNCTIONAL SUGAR
+
+In order to minimize typing, a less cumbersome prototyped interface is provided:
+
+    use MVC::Neaf qw(:sugar);
+
+    get '/foo/bar' => sub { ... }, view => 'TT';
+    neaf error => 404 => \&my_error_template;
+
+    neaf->run;
+
+It is not stable yet, so be careful when upgrading Neaf.
+
+=head2 get '/path' => CODE, %options;
+
+Create a route with C<GET/HEAD> methods enabled.
+The %options are the same as those of C<route()> method.
+
+=cut
+
+sub get(@) { ## no critic # DSL
+    my ($path, $handler, @args) = @_;
+
+    return MVC::Neaf->route( $path, $handler, method => [ 'GET', 'HEAD' ], @args );
+};
+
+=head2 post '/path' => CODE, %options;
+
+Create a route with C<POST> method enabled.
+The %options are the same as those of C<route()> method.
+
+=cut
+
+sub post(@) { ## no critic # DSL
+    my ($path, $handler, @args) = @_;
+
+    return MVC::Neaf->route( $path, $handler, method => [ 'POST' ], @args );
+};
+
+=head2 neaf->...
+
+Returns default Neaf instance (C<$MVC::Neaf::Inst>), so that
+C<neaf-E<gt>method_name> is the equivalent of C<MVC::Neaf-E<gt>method_name>.
+
+=head2 neaf shortcut => @options;
+
+Shorter alias to method described above. Currently supported:
+
+=over
+
+=item * route - C<route>
+
+=item * error - C<set_error_handler>
+
+=item * view - C<load_view>
+
+=back
+
+=cut
+
+my %method_shortcut = (
+    route => 'route',
+    error => 'set_error_handler',
+    view  => 'load_view',
+);
+
+sub neaf(@) { ## no critic # DSL
+    return $MVC::Neaf::Inst unless @_;
+
+    my ($action, @args) = @_;
+
+    my $method = $method_shortcut{$action};
+    croak "neaf: don't know how to handle '$action'"
+        unless $method;
+
+    return MVC::Neaf->$method( @args );
+};
+
+=head1 EXPERIMENTAL HOOKS
+
+Hooks are subroutines executed during various phases of request processing.
+Each hook is characterized by phase, code to be executed, path, and method.
+Multiple hooks MAY be added for the same phase/path/method combination.
+ALL hooks matching a given route will be executed, either short to long or
+long to short (aka "event bubbling"), depending on the phase.
+
+B<CAUTION> Avoid hooks if possible.
+Hook handling code is the most convoluted, complex part of the Neaf CORE.
+
+Use normal subroutine calls,
+even if copied-and-pasted across multiple controllers.
+Use guard objects if you want automatic resource deallocation.
+Subclass L<MVC::Neaf::View> and/or its children if you want to add an action
+around content rendering.
+
+Only use hooks when other options are exhausted,
+or there's a clear, straightforward use case.
+
+=head2 add_hook ( phase => CODEREF, %options )
+
+Set execution hook for given phase. See list of phases below.
+
+The CODEREF receives one and only argument - the C<$request> object.
+Return value is ignored.
+
+Use the following primitives to maintain state accross hooks and the main
+controller:
+
+=over
+
+=item * Use C<session> if you intend to share data between requests.
+
+=item * Use C<reply> if you intend to render the data for the user.
+
+=item * Use C<stash> as a last resort for temporary, private data.
+
+=back
+
+%options may include:
+
+=over
+
+=item * path => '/path' - where the hook applies. Default is '/'.
+Multiple locations may be supplied via C<[ /foo, /bar ...]>
+
+=item * exclude => '/path/dont' - don't apply to these locations,
+even if under '/path'.
+Multiple locations may be supplied via C<[ /foo, /bar ...]>
+
+=item * method => 'METHOD' || [ list ]
+List of request HTTP methods to which given hook applies.
+
+=back
+
+=head2 HOOK PHASES
+
+This list of phases MAY change in the future.
+
+=head3 pre_logic
+
+Executed AFTER finding the correct route, but BEFORE processing the main
+handler code (one that returns C<\%hash>, see C<route> above).
+
+Hooks are executed in order, shorted paths to longer.
+C<reply> is not available at this stage,
+as the controller has not been executed yet.
+
+Dying in this phase stops both further hook processing and controller execution.
+Instead, the corresponding error handler is executed right away.
+
+B<EXAMPLE>: use this hook to produce a 403 error if the user is not logged in
+and looking for a restricted area of the site:
+
+    MVC::Neaf->set_hook( pre_logic => sub {
+        my $request = shift;
+        $request->session->{user_id} or die 403;
+    }, path => '/admin', exclude => '/admin/static' );
+
+=head3 pre_content
+
+This hook is run AFTER the main handler has returned or died, but BEFORE
+content rendering/serialization is performed.
+
+C<reply()> hash is available at this stage.
+
+Dying is ignored, only producing a warning.
+
+=head3 pre_reply
+
+This hook is run AFTER the headers have been generated, but BEFORE the reply is
+actually sent to client. This is the last chance to amend something.
+
+Hooks are executed in REVERSE order, from longer to shorter paths.
+
+C<reply()> hash is available at this stage.
+
+Dying is ignored, only producing a warning.
+
+=head3 pre_cleanup
+
+This hook is run AFTER all postponed actions set up in controller
+(via C<-continue> etc), but BEFORE the request object is actually destroyed.
+This can be useful to deinitialize something or write statistics.
+
+The client conection MAY be closed at this point and SHOULD NOT be relied upon.
+
+Hooks are executed in REVERSE order, from longer to shorter paths.
+
+Dying is ignored, only producing a warning.
+
+=cut
+
+my %add_hook_args;
+$add_hook_args{$_}++ for qw(method path exclude);
+my %hook_phases;
+$hook_phases{$_}++ for qw(pre_logic pre_content pre_reply pre_cleanup);
+
+sub add_hook {
+    my ($self, $phase, $code, %opt) = @_;
+    $self = $Inst unless ref $self;
+
+    my @extra = grep { !$add_hook_args{$_} } keys %opt;
+    $self->_croak( "unknown options: @extra" )
+        if @extra;
+    $self->_croak( "illegal phase: $phase" )
+        unless $hook_phases{$phase};
+
+    _listify( \$opt{method}, qw( GET HEAD POST PUT PATCH DELETE ) );
+    _listify( \$opt{path}, '/' );
+    _listify( \$opt{exclude} );
+    @{ $opt{path} } = map { canonize_path($_) } @{ $opt{path} };
+
+    $opt{caller} = [ caller(0) ]; # where the hook was set
+    $opt{phase}  = $phase; # just for information
+    $opt{code}   = $code;
+
+    # hooks == {method}{path}{phase}[nnn] => { code => CODE, ... }
+
+    foreach my $method ( @{$opt{method}} ) {
+        foreach my $path ( @{$opt{path}} ) {
+            my $where = $self->{hooks}{$method}{$path}{$phase} ||= [];
+            push @$where, \%opt;
+        };
+    };
+
+    return $self;
+};
+
+# TODO util?
+# usage: listify ( \$var, default1, default2... )
+# converts scalar in-place to arrayref if needed
+sub _listify {
+    my ($scalref, @default) = @_;
+
+    if (ref $$scalref ne 'ARRAY') {
+        my $array = defined $$scalref ? [ my $tmp = $$scalref ] : \@default;
+        $$scalref = $array;
+    };
+
+    return $$scalref;
+};
+
 =head1 DEVELOPMENT AND DEBUGGING METHODS
 
 =head2 get_routes
@@ -873,7 +1158,7 @@ But it's possible.
 
 Options are not checked whatsoever.
 
-Just in case you're curious, $MVC::Neaf::Inst is the default instance
+Just in case you're curious, C<$MVC::Neaf::Inst> is the default instance
 that handles MVC::Neaf->... requests.
 
 =cut
@@ -883,29 +1168,27 @@ sub new {
 
     $opt{-type}     ||= "text/html";
     $opt{-view}     ||= "TT";
-    $opt{defaults}  ||= { -status => 200 };
-    $opt{on_error}  ||= sub {
-        my ($req, $err, $where) = @_;
-        my $msg = "ERROR: ".$req->script_name.": $err";
-        if ($where) {
-            $msg =~ s/\s+$//s;
-            $msg .= " in $where->[1] line $where->[2]";
-        };
-        warn "$msg\n";
-    };
     my $force = delete $opt{force_view};
 
     my $self = bless \%opt, $class;
+
+    $self->{on_error} ||= sub {
+        my ($req, $err, $where) = @_;
+        $self->_log_error( $req->script_name, $err );
+    };
+
     $self->set_forced_view( $force )
         if $force;
+    # avoid an extra ||= in handle requ
+    $self->set_path_defaults( '/' => { -status => 200 } );
 
     return $self;
 };
 
-=head2 handle_request( MVC::Neaf::request->new )
+=head2 handle_request( MVC::Neaf::Request->new )
 
 This is the CORE of this module.
-Should not be called directly - use run() instead.
+Should not be called directly - use C<run()> instead.
 
 =cut
 
@@ -940,7 +1223,11 @@ sub handle_request {
         $path_info_regex =~ $route->{path_info_regex_regex}
             or die "404\n";
         $req->set_full_path( $path, $path_info_regex, !$route->{has_path_info_regex} );
+        $self->_post_setup( $route )
+            unless exists $route->{lock};
 
+        # execute hooks
+        $_->($req) for @{ $route->{hooks}{pre_logic} };
         # Run the controller!
         return $route->{code}->($req);
     };
@@ -949,7 +1236,7 @@ sub handle_request {
         # post-process data - fill in request(RD) & global(GD) defaults.
         # TODO fill in per-location defaults, too - but do we need them?
         my $RD = $req->get_default;
-        my $GD = $self->{defaults};
+        my $GD = $route->{default} ||= $self->_make_defaults( $route );
         exists $data->{$_} or $data->{$_} = $RD->{$_} for keys %$RD;
         exists $data->{$_} or $data->{$_} = $GD->{$_} for keys %$GD;
     } else {
@@ -959,44 +1246,55 @@ sub handle_request {
 
     # END ROUTE REQUEST
 
+    $req->_set_reply( $data );
     exists $self->{stat}
         and $self->{stat}->record_controller($req->script_name);
+    if (exists $route->{hooks}{pre_content}) {
+        foreach my $hook ( @{ $route->{hooks}{pre_content} } ) {
+            if (!eval { $hook->( $req ); 1 }) {
+                $self->_log_error( "pre_content hook", $@ );
+            };
+        };
+    };
 
     # PROCESS REPLY
 
     # Render content if needed. This may alter type, so
     # produce headers later.
-    my ($content, $type);
-    if (defined $data->{-content}) {
-        $content = $data->{-content};
-    } else {
+    my $content = \$data->{-content};
+    if( !defined $$content) {
         my $view = $self->load_view( $data->{-view} || $route->{view} );
+        # TODO move session to hook
         if ( my $key = $self->{session_view_as} and my $sess = $req->session(1) ) {
             $data->{ $key } = $sess;
         };
-        eval { ($content, $type) = $view->render( $data ); };
-        if (!defined $content) {
-            warn "ERROR: In view: $@";
+        eval {
+            ($$content, my $type) = $view->render( $data );
+            $data->{-type} ||= $type;
+        };
+        if (!defined $$content) {
+            # TODO $req->clear in case of error
+            $self->_log_error( view => $@ );
             $data = {
                 -status => 500,
                 -type   => "text/plain",
             };
-            $content = "Template error.";
+            $$content = "Template error.";
         };
-        $data->{-type} ||= $type || 'text/html';
     };
 
+    # TODO should this be a sub?
     # Encode unicode content NOW so that we don't lie about its length
     # Then detect ascii/binary
-    if (Encode::is_utf8( $content )) {
+    if (Encode::is_utf8( $$content )) {
         # UTF8 means text, period
-        $content = encode_utf8( $content );
+        $$content = encode_utf8( $$content );
         $data->{-type} ||= 'text/plain';
         $data->{-type} .= "; charset=utf-8"
             unless $data->{-type} =~ /; charset=/;
     } elsif (!$data->{-type}) {
         # Autodetect binary. Plain text is believed to be in utf8 still
-        $data->{-type} = $content =~ /^.{0,512}?[^\s\x20-\x7F]/s
+        $data->{-type} = $$content =~ /^.{0,512}?[^\s\x20-\x7F]/s
             ? 'application/octet-stream'
             : 'text/plain; charset=utf-8';
     } elsif ($data->{-type} =~ m#^text/#) {
@@ -1011,30 +1309,108 @@ sub handle_request {
     $head->init_header( location => $data->{-location} )
         if $data->{-location};
     $head->push_header( set_cookie => $req->format_cookies );
-    $head->init_header( content_length => length $content )
+    $head->init_header( content_length => length $$content )
         unless $data->{-continue};
     $head->init_header( expires => http_date( time + $route->{cache_ttl} ) )
         if exists $route->{cache_ttl} and $data->{-status} == 200;
-    $content = '' if $req->method eq 'HEAD';
+    $$content = '' if $req->method eq 'HEAD';
 
     # END PROCESS REPLY
 
+    if (exists $route->{hooks}{pre_reply}) {
+        foreach my $hook ( @{ $route->{hooks}{pre_reply} } ) {
+            if (!eval { $hook->( $req ); 1 }) {
+                $self->_log_error( "pre_reply hook", $@ );
+            };
+        };
+    };
     exists $self->{stat}
         and $self->{stat}->record_finish($data->{-status}, $req);
+    if (exists $route->{hooks}{pre_cleanup}) {
+        $req->postpone( $route->{hooks}{pre_cleanup} );
+    };
 
     # DISPATCH CONTENT
 
     if ($data->{-continue} and $req->method ne 'HEAD') {
         $req->postpone( $data->{'-continue'}, 1 );
-        $req->postpone( sub { $_[0]->write( $content ); }, 1 );
+        $req->postpone( sub { $_[0]->write( $$content ); }, 1 );
         return $req->do_reply( $data->{-status} );
     } else {
-        return $req->do_reply( $data->{-status}, $content );
+        return $req->do_reply( $data->{-status}, $$content );
     };
 
     # END DISPATCH CONTENT
 }; # End handle_request()
 
+# _post_setup( $route )
+# 1) calc defaults
+# 2) calc hooks
+# 3) lock
+sub _post_setup {
+    my ($self, $route) = @_;
+
+    # CALCULATE DEFAULTS
+    my %def;
+    # merge data sources, longer paths first
+    my @sources = (
+          $route->{todo_default}
+        , map { $self->{path_defaults}{$_} }
+            reverse path_prefixes( $route->{path} )
+    );
+
+    foreach my $src( @sources ) {
+        $src or next;
+        exists $def{$_} or $def{$_} = $src->{$_}
+            for keys %$src;
+    };
+
+    # kill undef values
+    defined $def{$_} or delete $def{$_}
+        for keys %def;
+    $route->{default} = \%def;
+
+    # CALCULATE HOOKS
+    # select ALL hooks prepared for upper paths
+    my $hook_tree = $self->{hooks}{ $route->{my_method} };
+    my @hook_by_path =
+        map { $hook_tree->{$_} || () } path_prefixes( $route->{path} );
+
+    # Merge callback stacks into one hash, in order
+    # hook = {method}{path}{phase}[nnn] => { code => sub{}, ... }
+    # We need to extract that sub {}
+    # We do so in a rather clumsy way that would short cirtuit
+    #     at all possibilities
+    # Premature optimization FTW!
+    my %phases;
+    foreach my $hook_by_phase (@hook_by_path) {
+        foreach my $phase ( keys %$hook_by_phase ) {
+            my $hook_list = $hook_by_phase->{$phase};
+            foreach my $hook (@$hook_list) {
+                # process excludes - if path starts with any, no go!
+                grep { $route->{path} =~ m#^\Q$_\E(?:/|$)# }
+                    @{ $hook->{exclude} }
+                        and next;
+                # TODO filter out repetition
+                push @{ $phases{$phase} }, $hook->{code};
+                # TODO also store hook info somewhere for better error logging
+            };
+        };
+    };
+
+    # the pre-reply, pre-cleanup should go in backward direction
+    # those are for cleaning up stuff
+    $phases{$_} and @{ $phases{$_} } = reverse @{ $phases{$_} }
+        for qw(pre_cleanup pre_reply);
+
+    $route->{hooks} = \%phases;
+
+    # LOCK PROFILE
+    $route->{lock}++
+        and die "MVC::Neaf broken, please file a bug";
+    return;
+};
+
 sub _error_to_reply {
     my ($self, $req, $err, $where) = @_;
 
@@ -1046,12 +1422,9 @@ sub _error_to_reply {
     my $status = (!ref $err && $err =~ /^(\d\d\d)/) ? $1 : 500;
 
     # Try exception handler
-    if( !$1 ) {
-        exists $self->{on_error}
-            and eval { $self->{on_error}->($req, $err, $where) };
-        # ignore errors in error handler
-        warn "ERROR: Error handler failed: $@"
-            if $@;
+    if( !$1 and exists $self->{on_error}) {
+        eval { $self->{on_error}->($req, $err, $where); 1 }
+            or $self->_log_error( "error handler", $@ );
     };
 
     # Try fancy error template
@@ -1065,8 +1438,7 @@ sub _error_to_reply {
         };
         return $ret
             if (ref $ret eq 'HASH');
-        warn "ERROR: Error status handler for $status failed: $@"
-            if $@;
+        $self->_log_error( "status $status handler:", $@ );
     };
 
     # Options exhausted - return plain error message
@@ -1085,21 +1457,49 @@ sub _croak {
     croak( (ref $self || $self)."->$where: $msg" );
 };
 
+sub _log_error {
+    my ($self, $where, $err) = @_;
+
+    my $msg = "ERROR: in $where: $err";
+    $msg =~ s/\n\s*/ /gs;
+    $msg =~ s/\s*$/\n/;
+    warn $msg;
+};
+
 =head2 run_test( \%PSGI_ENV )
 
-Run a PSGI request and return ($status, HTTP::Headers, $whole_content ).
+=head2 run_test( "/path?param=value" )
+
+Run a PSGI request and return a list of
+C<($status, HTTP::Headers, $whole_content )>.
 
 Just as the name suggests, useful for testing only (it reduces boilerplate).
 
-Does NOT handle continuation requests.
+Continuation responses are supported.
 
 =cut
 
 sub run_test {
     my ($self, $env) = @_;
 
+    if (!ref $env) {
+        $env =~ /^(.*?)(?:\?(.*))?$/;
+        $env = {
+            REQUEST_URI => $env,
+            REQUEST_METHOD => 'GET',
+            QUERY_STRING => defined $2 ? $2 : '',
+            SERVER_NAME => 'localhost',
+            SERVER_PORT => 80,
+            SCRIPT_NAME => '',
+            PATH_INFO => $1,
+        }
+    };
     my $ret = $self->run->( $env );
-    # TODO handle functional return, too!
+    if (ref $ret eq 'CODE') {
+        # PSGI functional interface used.
+        require MVC::Neaf::Request::FakeWriter;
+        $ret = MVC::Neaf::Request::FakeWriter->new->respond( $ret );
+    };
 
     return (
         $ret->[0],
