@@ -3,7 +3,7 @@ package MVC::Neaf::Request;
 use strict;
 use warnings;
 
-our $VERSION = 0.1302;
+our $VERSION = 0.14;
 
 =head1 NAME
 
@@ -16,7 +16,8 @@ This is what your L<MVC::Neaf> application is going to get as its ONLY input.
 Here's a brief overview of what a Neaf request returns:
 
     # How the application was configured:
-    MVC::Neaf->route( "/matching/route" => sub { my $req = shift; ... } );
+    MVC::Neaf->route( "/matching/route" => sub { my $req = shift; ... },
+        path_info_regex => '.*' );
 
     # What was requested:
     http(s)://server.name:1337/mathing/route/some/more/slashes?foo=1&bar=2
@@ -29,8 +30,9 @@ Here's a brief overview of what a Neaf request returns:
     $req->port         = 1337
     $req->path         = /mathing/route/some/more/slashes
     $req->script_name  = /mathing/route
+    $req->path_info    = some/more/slashes
 
-    $req->path_info( '.+' )     = some/more/slashes
+    # params and cookies require a regexp
     $req->param( foo => '\d+' ) = 1
 
 =head1 REQUEST METHODS
@@ -139,7 +141,7 @@ sub method {
 
 =head2 is_post()
 
-Alias for $self->method eq 'POST'.
+Alias for C<$self-E<gt>method eq 'POST'>.
 May be useful in form submission, as in
 
     $form = $request->form( $validator );
@@ -199,7 +201,7 @@ sub path {
 
 The part of the request that mathed the route to the
 application being executed.
-Guaranteed to start with slash and be a prefix of path().
+Guaranteed to start with slash and be a prefix of C<path()>.
 
 =cut
 
@@ -271,6 +273,7 @@ sub set_full_path {
         $script_name = $self->do_get_path;
     };
 
+    # CANONIZE
     $script_name =~ s#^/*#/#;
     $self->{script_name} = $script_name;
 
@@ -306,6 +309,7 @@ sub set_path_info {
     my ($self, $path_info) = @_;
 
     $path_info = '' unless defined $path_info;
+    # CANONIZE
     $path_info =~ s#^/+##;
 
     $self->{path_info} = $path_info;
@@ -564,6 +568,8 @@ sub set_default {
 
 Returns a hash of previously set default values.
 
+B<EXPERIMANTAL>. API and naming subject to change.
+
 =cut
 
 sub get_default {
@@ -666,7 +672,7 @@ sub set_cookie {
         $opt{expires} = time + $opt{ttl};
     };
 
-    $self->{neaf_cookie_out}{ $name } = [
+    $self->{response}{cookie}{ $name } = [
         $cook, $opt{regex},
         $opt{domain}, $opt{path}, $opt{expires}, $opt{secure}, $opt{httponly}
     ];
@@ -704,7 +710,7 @@ ready to be put into Set-Cookie header.
 sub format_cookies {
     my $self = shift;
 
-    my $cookies = $self->{neaf_cookie_out} || {};
+    my $cookies = $self->{response}{cookie} || {};
 
     my @out;
     foreach my $name (keys %$cookies) {
@@ -1073,7 +1079,7 @@ B<NOTE> This format may change in the future.
 sub header_out {
     my $self = shift;
 
-    my $head = $self->{header_out} ||= HTTP::Headers->new;
+    my $head = $self->{response}{header} ||= HTTP::Headers->new;
     return $head unless @_;
 
     my $name = shift;
@@ -1108,9 +1114,59 @@ sub remove_header {
     return $self->header_out->remove_header( $name );
 };
 
+=head2 reply
+
+Returns reply hashref that was returned by controller, if any.
+Returns undef unless the controller was actually called.
+This may be useful in postponed actions or hooks.
+
+This is killed by a C<clear()> call.
+
+B<EXPERIMENTAL>. This function MAY be removed or changed in the future.
+
+=cut
+
+sub reply {
+    my $self = shift;
+
+    return $self->{response}{ret};
+}
+
+sub _set_reply {
+    my ($self, $data) = @_;
+    $self->{response}{ret} = $data;
+    return $self;
+}
+
+=head2 stash()
+
+A hashref that is guaranteed to persist throughout the request lifetime.
+
+This may be useful to maintain shared data accross hooks and callbacks.
+
+Use C<session> if you intend to share data between requests.
+Use C<reply> if you intend to render the data for the user.
+Use C<stash> as a last resort for temporary, private data.
+
+Stash is not killed by C<clear()> function so that cleanup isn't
+botched accidentally.
+
+B<EXPERIMENTAL>. This function MAY be removed if hooks turn out to be
+too cumbersome.
+
+=cut
+
+sub stash {
+    my $self = shift;
+    my $st = $self->{stash} ||= {};
+    return $st;
+};
+
 =head2 postpone( CODEREF->(req) )
 
-Execute a function right after the request is served.
+=head2 postpone( [ CODEREF->(req), ... ] )
+
+Execute a function (or several) right after the request is served.
 
 Can be called multiple times.
 
@@ -1128,16 +1184,14 @@ Returns self.
 sub postpone {
     my ($self, $code, $prepend_flag) = @_;
 
-    ref $code eq 'CODE'
-        or $self->_croak( "argument must be a function" );
+    $code = [ $code ]
+        unless ref $code eq 'ARRAY';
+    grep { ref $_ ne 'CODE' } @$code
+        and $self->_croak( "argument must be a function or a list of functions" );
 
     $prepend_flag
-        ? unshift @{ $self->{postponed} }, $code
-        : push    @{ $self->{postponed} }, $code;
-
-    # TODO UGLY HACK - rely on side-effect to determine state
-    $self->{continue}++;
-    $self->{buffer_size} ||= 4096;
+        ? unshift @{ $self->{response}{postponed} }, reverse @$code
+        : push    @{ $self->{response}{postponed} }, @$code;
 
     return $self;
 };
@@ -1179,6 +1233,24 @@ sub close {
     return $self->do_close();
 }
 
+=head2 clear()
+
+Remove all data that belongs to reply.
+This is called when a handler bails out to avoid e.g. setting cookies
+in a failed request.
+
+=cut
+
+sub clear {
+    my $self = shift;
+
+    $self->_croak( "called after responding" )
+        if $self->{continue};
+
+    delete $self->{response};
+    return $self;
+}
+
 =head1 METHODS FOR DRIVER DEVELOPERS
 
 The following methods are to be used/redefined by backend writers
@@ -1200,11 +1272,12 @@ Returns self.
 sub execute_postponed {
     my $self = shift;
 
-    my $todo = delete $self->{postponed};
+    $self->{continue}++;
+    my $todo = delete $self->{response}{postponed};
     foreach my $code (@$todo) {
         # avoid dying in DESTROY, as well as when serving request.
         eval { $code->($self); };
-        print STDERR "WARN ".(ref $self).": postponed action failed: $@"
+        carp "WARN ".(ref $self).": postponed action failed: $@"
             if $@;
     };
 
@@ -1215,7 +1288,7 @@ sub DESTROY {
     my $self = shift;
 
     $self->execute_postponed
-        if (exists $self->{postponed});
+        if (exists $self->{response}{postponed});
 };
 
 =head1 DRIVER METHODS
